#!/usr/bin/python
# _param_help.py

# flake8: noqa


'''
Helper functions for assessing the effects of different parameter settings.
'''

# geonomics imports
from geonomics.utils.viz import _check_display

# other imports
import os
import numpy as np
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from scipy.stats import levy as _s_levy
from copy import deepcopy

# ------------------------------------
# CLASSES ---------------------------
# ------------------------------------


# --------------------------------------
# FUNCTIONS ---------------------------
# --------------------------------------



def plot_movement(spp, land, num_timesteps, lyr_num=None, params=None,
                  direction_distr_mu=None, direction_distr_kappa=None,
                  distance_distr_param1=None, distance_distr_param2=None,
                  distance_distr='levy',
                  move_surf=None, subset_spp=None, color='black',
                  alpha=None, land_cmap='plasma',  size=10, ticks=False,
                  color_by_individ=True, increasing_linewidth=True,
                  include_start_points=False):

    '''
    Plot example movement tracks generated by the input movement parameters

    Input a ParametersDict object with movement parameter values of interest,
    or input those values in piecemeal fashion, to generate a plot of example
    movement tracks generated by that parameterization. Useful for visual
    exploration of the movement trends in a Geonomics model.

    Parameters
    ----------
    spp : :class:`geonomics.structs.Species`
        A Species object whose movement behavior should be plotted

    land : :class:`geonomics.structs.Landscape`
        A Land object upon which the Species' movement should be plotted

    num_timesteps : int
        A number of timesteps (i.e. individual movement steps) to be plotted in
        each individual's movement track.

    lyr_num : int, optional, default: None
        The number of the Layer to be plotted (i.e. its key in the Landscape
        dict). If None, all Layers will be plotted as a stack of transparent,
        differentially colormapped rasters.

    params : {:class:`geonomics.sim.params.ParametersDict`}, optional,
    default: None
        If a ParametersDict object is passed, then the movement resulting
        from the movement parameter values it contains will be plotted (except
        for any values that are manually overridden by new values passed
        individually to any of the following four parameters). If None is
        passed instead, then all four of the following movement parameters'
        values must be passed, and their resulting movement behavior will be
        plotted.

    direction_distr_mu,
    direction_distr_kappa,
    distance_distr_param1,
    distance_distr_param2 : float, optional default: None
        These are the parameters defining the distributions from which the
        direction and distance values for individual movements are drawn. If
        passed, their values will be used to determine the movement behavior
        that is plotted (even if a ParametersDict was passed to *params*,
        because these values will override any values therein). For more
        details, see the 'Parameters' section of the online documentation. 

    distance_distr: str, optional default: 'levy'
        This indicates the distribution to use for drawing movement
        distancs. Defaults to 'levy', but can also be set to 'wald'.

    move_surf : :class:`geonomics.utils.spatial._ConductanceSurface`, optional,
    default: None
        If a _ConductanceSurface (i.e. a Species' movement or dispersal
        surface) is passed, then it will be used to determine the plotted
        movement behavior, instead of any movement or dispersal surface already
        associated with the Species object passed to *spp*.

    subset_spp: int, optional, default: None
        The number of Individuals to subset from the Species and plot (rather
        than plotting movement tracks for all Individuals in the Species, which
        can create a jumbled plot that is hard to inspect).

    color : valid mpl.plt color value, optional, default: 'black'
        Face color for the lines in the movement-track plot.  Passed to the
        *c* argument of matplotlib.pyplot.scatter.

    alpha : valid mpl.plt alpha value, optional, default: None
        Transparency value for the lines in the movement-track plot. Passed
        to the *alpha* argument of matplotlib.pyplot.scatter.

    size : scalar or array_like, optional, default: 10
        Size of the scatter points. Passed to the *s* argument of
        matplotlib.pyplot.scatter.

    land_cmap : {valid string, None}, optional, default: 'plasma'
        Colormap to use for plotting Layer rasters. If None, will default
        to the colormap automatically assigned to the Layer by Geonomics
        based on the Layer's index number in the Landscape dict. Can be
        passed any string that references a colormap in matplotlib.pyplot.cm.

    ticks : bool, optional, default: False
        If True, x- and y-axis ticks will be added to the plot.

    color_by_individ : bool, optional, default: True
        If True, each plotted Individual's movement track will have its own
        distinct color, making it easier to follow the track and distinguish it
        from others in the plot.

    increasing_linewidth : bool, optional, default: True
        If True, each Individual's movement track will get wider with each
        successive movement step, making it easier to distinguish
        directionality in the movement track.

    include_start_points : bool, optional, default: False
        If True, a scatter of points will be plotted to signify the starting
        points of individuals' movement tracks.
    '''

    #assert that either a params dict was provided or parameters were
    #stipulated
    assert (params is not None
        or (distance_distr_param1 is not None
            and distance_distr_param2 is not None
            and direction_distr_mu is not None
            and direction_distr_kappa is not None)), ("Either a parameters "
        "dictionary must be provided, or parameters must be individually "
        "defined as arguments.")

    #if a params dictionary was fed in, copy it as toy_params
    if params is not None:
        toy_params = deepcopy(params)
        #copy the movement params from toy_params
        toy_move_params = toy_params.comm.species[spp.name].movement
        #then override existing values in the movement params dictionary,
        #if they have new values that have been entered for this trial
        if direction_distr_mu != None:
            toy_move_params['direction_distr_mu'] = direction_distr_mu
        if direction_distr_kappa != None:
            toy_move_params['direction_distr_kappa'] = direction_distr_kappa
        if distance_distr_param1  != None:
            toy_move_params['movement_distance_distr_param1'] = distance_distr_param1
        if distance_distr_param2 != None:
            toy_move_params['movement_distance_distr_param2'] = distance_distr_param2
        if move_surf != None:
            toy_move_params['move_surf'] = move_surf
        toy_params.comm.species[spp.name].movement = toy_move_params

    #if no params dictionary provided, create toy_params from
    #the individual parameter values that were fed in
    elif params is None:
        toy_move_params = {}
        toy_move_params['direction_distr_mu'] = direction_distr_mu
        toy_move_params['direction_distr_kappa'] = direction_distr_kappa
        toy_move_params['movement_distance_distr_param1'] = distance_distr_param1
        toy_move_params['movement_distance_distr_param2'] = distance_distr_param2
        if move_surf != None:
            toy_move_params['move_surf'] = move_surf
        elif spp.move_surf == None:
            toy_move_params['move_surf'] = False
        else:
            toy_move_params['move_surf'] = True
            toy_params = {'comm':{'species':{'movement':toy_move_params}}}

    #now copy the species to a toy object
    toy_spp = deepcopy(spp)
    #replace its movement-parameter values
    for param in ['direction_distr_mu', 'direction_distr_kappa',
        'movement_distance_distr_param1', 'movement_distance_distr_param2']:
        setattr(toy_spp, param,
            toy_params['comm']['species'][spp.name]['movement'][param])

    #replace its movement surface, if necessary
    if move_surf is not None:
        toy_spp._move_surf = move_surf

    #subset individuals, if necessary
    if subset_spp:
        cull_individs = np.random.choice([*toy_spp],
            len(toy_spp) - subset_spp, replace = False)
        [toy_spp.pop(ind) for ind in cull_individs]
        #reset the coords before plotting
        toy_spp._set_coords_and_cells()

    #set the plotting linewidths to increase over runtime
    if increasing_linewidth:
        linewidths = np.linspace(1, 5, num=num_timesteps)
    else:
        linewidths = np.array([1] * num_timesteps)

    #Create a colors list, to plot different-colored paths for individuals
    #colors = ['black', 'red', 'orange', 'yellow', 'green', 'blue', 'white']
    #colors = [plt.cm.Greys_r(i) for i in np.linspace(0, 0.9, 9)]
    colors = ['#000000', '#2F4F4F', '#778899', '#696969', '#A9A9A9', '#C0C0C0']

    # get list of order in which to plot individuals
    ind_idx_list = [i.idx for i in toy_spp.values()]
    ind_idx_list = sorted(ind_idx_list)

    #plot the species at its starting locations
    ind_colors = [colors[i % len(colors)] for i in range(len(ind_idx_list))]
    toy_spp._plot(lyr_num=lyr_num, land=land, color=ind_colors,
                  size=25 * include_start_points, land_cmap=land_cmap,
                  ticks=ticks)

    #set the new_x and new_y objects to the current locations, before for loop
    #that will iteratively update them to the newer locations after movement
    new_x = [getattr(toy_spp[idx], 'x') for idx in ind_idx_list]
    new_y = [getattr(toy_spp[idx], 'y') for idx in ind_idx_list]

    #loop for the number of timesteps, iteratively moving individuals and
    #plotting the vectors between their previous and new positions
    for t in range(num_timesteps):
        old_x = [x for x in new_x]
        old_y = [y for y in new_y]

        toy_spp._do_movement(land)

        new_x = [getattr(toy_spp[idx], 'x') for idx in ind_idx_list]
        new_y = [getattr(toy_spp[idx], 'y') for idx in ind_idx_list]

        #plot with separate colors for each individual, if necessary
        if color_by_individ:
            [plt.plot((old_x[i], new_x[i]), (old_y[i], new_y[i]), '-',
                scalex = False, scaley = False, linewidth = linewidths[t],
                color = colors[i%len(colors)],
                alpha = 0.5) for i in range(len(old_x))];

        #or else plot without individuals colored differently
        else:
            plt.plot((old_x, new_x), (old_y, new_y), '-',
                scalex = False, scaley = False, linewidth = linewidths[t],
                color = color, alpha = alpha)


# functions to visualize the different distributions that can be
# parameterized from the parameters file
def plot_distr_movement_distance(spp=None, distance_distr_param1=None,
                                 distance_distr_param2=None,
                                 distance_dist='levy'):
    if spp is not None:
        distance_distr_param1 = spp.movement_distance_distr_param1
        distance_distr_param2 = spp.movement_distance_distr_param2
    else:
        assert (distance_distr_param1 is not None and
                distance_distr_param2 is not None), ('If a Species object '
                                                     'is not provided then '
                                                     'the parameter values '
                                                     'must be provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if distance_distr == 'levy':
        fig.suptitle(('movement distance: '
                  '~Levy($\loc$=%.4E, $\scale$=%.4E)') % (distance_distr_param1,
                                                        distance_distr_param2))
        vals = _s_levy.rvs(loc=distance_distr_param1,
                              scale=distance_distr_param2, size=10000)
    elif distance_distr == 'wald':
        fig.suptitle(('movement distance: '
                  '~Wald($\mu$=%.4E, $\scale$=%.4E)') % (distance_distr_param1,
                                                         distance_distr_param2))
        vals = np.random.wald(mean=distance_distr_param1,
                              scale=distance_distr_param2, size=10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()


def plot_distr_movement_direction(spp=None, direction_distr_mu=None,
                            direction_distr_kappa=None):
    if spp is not None:
        direction_distr_mu = spp.direction_distr_mu
        direction_distr_sigma = spp.direction_distr_kappa
    else:
        assert (direction_distr_mu is not None and
                direction_distr_kappa is not None), ('If a Species object '
                                                     'is not provided then '
                                                     'the parameter values '
                                                     'must be provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig.suptitle(('movement direction: '
                 '~von Mises($\mu$ = %.4E, '
                  '$\kappa$ = %.4E)') % (direction_distr_mu,
                                         direction_distr_kappa))
    vals = np.random.vonmises(direction_distr_mu, direction_distr_kappa, 10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()


def plot_distr_dispersal_distance(spp=None, dispersal_distr_param1=None,
                                  dispersal_distr_param2=None,
                                  dispersal_distance_distr='levy'):
    if spp is not None:
        dispersal_distr_param1 = spp.dispersal_distance_distr_param1
        dispersal_distr_param2 = spp.dispersal_distance_distr_param2
    else:
        assert (dispersal_distr_param1 is not None and
                dispersal_distr_param2 is not None), ('If a Species object '
                                                     'is not provided then '
                                                     'the parameter values '
                                                     'must be provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if dispersal_distance_distr == 'levy':
        fig.suptitle(('dispersal distance: '
                      '~Levy($\loc$=%.4E, $\scale$=%.4E)') % (dispersal_distr_param1,
                                                           dispersal_distr_param2))
        vals = _s_levy.rvs(dispersal_distr_param1, dispersal_distr_param2, 10000)
    elif dispersal_distance_distr == 'wald':
        fig.suptitle(('dispersal distance: '
                      '~Wald($\mean$=%.4E, $\scale$=%.4E)') % (dispersal_distr_param1,
                                                           dispersal_distr_param2))
        vals = np.random.wald(dispersal_distr_param1, dispersal_distr_param2, 10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()


def plot_n_births(spp=None, n_births_distr_lambda=None):
    if spp is not None:
        n_births_distr_lambda = spp.n_births_distr_lambda
    else:
        assert n_births_distr_lambda is not None ('If a Species object '
                                                  'is not provided then '
                                                  'the parameter values '
                                                  'must be provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig.suptitle(('number of births per mating event: '
                  '~Poisson($\lamba$=%.4E)') % n_births_distr_lambda)
    vals = np.random.poisson(n_births_distr_lambda, 10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()


def plot_distr_delet_effect_sizes(spp=None, delet_alpha_distr_shape=None,
                                  delet_alpha_distr_scale=None):
    if spp is not None:
        delet_alpha_distr_shape = spp.gen_arch.delet_alpha_distr_shape
        delet_alpha_distr_scale = spp.gen_arch.delet_alpha_distr_scale
    else:
        assert (delet_alpha_distr_shape is not None
                and delet_alpha_distr_scale is not None), ('If a Species '
                                                           'object is not '
                                                           'provided then '
                                                           'the parameter '
                                                           'values must be '
                                                           'provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig.suptitle(('effect sizes of deleterious mutations: '
                  '~Gamma($shape$=%.4E, '
                  '$scale$=%.4E)') % (delet_alpha_distr_shape,
                                      delet_alpha_distr_scale))
    vals = np.random.gamma(delet_alpha_distr_shape,
                           delet_alpha_distr_scale, 10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()


def plot_distr_trait_effect_sizes(spp=None, r_distr_alpha=None,
                            r_distr_beta=None):
    if spp is not None:
        r_distr_alpha = spp.gen_arch.recombinations._r_distr_alpha
        r_distr_beta = spp.gen_arch.recombinations._r_distr_beta
    else:
        assert (r_distr_alpha is not None
                and r_distr_beta is not None), ('If a Species '
                                                'object is not '
                                                'provided then '
                                                'the parameter '
                                                'values must be '
                                                'provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig.suptitle(('recombination rates: '
                  '~Beta($\alpha$=%.4E, $\beta$=%.4E)') % (r_distr_alpha,
                                                           r_distr_beta))
    vals = np.random.beta(r_distr_alpha, r_distr_beta, 10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()


def plot_distr_recomb_rates(spp=None, trt_num=None, alpha_distr_mu=None,
                            alpha_distr_sigma=None):
    if spp is not None:
        assert trt_num is not None, ('Trait number must be provided if a '
                                     'Species object is provided.')
        alpha_distr_mu = spp.gen_arch.traits[trt_num]._alpha_distr_mu
        alpha_distr_sigma = spp.gen_arch.traits[trt_num]._alpha_distr_sigma
    else:
        assert (alpha_distr_mu is not None
                and alpha_distr_sigma is not None), ('If a Species '
                                                     'object is not '
                                                     'provided then '
                                                     'the parameter '
                                                     'values must be '
                                                     'provided.')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig.suptitle(('effect sizes of loci, trait %i: '
                  '~Normal($\mu$=%.4E, $\sigma$=%.4E)') % (alpha_distr_mu,
                                                           alpha_distr_sigma))
    vals = np.random.normal(alpha_distr_mu, alpha_distr_sigma, 10000)
    ax.hist(vals, bins = 25)
    ax.set_xlim((min(vals), max(vals)))
    plt.show()



#!/usr/bin/python
# mating.py

# flake8: noqa

'''
Functions to implement mating operations.
'''

#other imports
from scipy.spatial import cKDTree
import numpy as np
import numpy.random as r
from operator import itemgetter as ig
from itertools import repeat, starmap


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

def _find_mates(spp, sex=False, repro_age=None, choose_nearest=False,
                inverse_dist_mating=False):
    # NOTE: In an IDEAL world, this would work as follows:
    # for all individuals, find set of nearest neighbors (excluding selves)
    # for females:
        # choose male from among those nn's, based on:
            # probabilistic function of distance
        # if sexual_selection:
            # (product of?) selection coefficients
    ######################################################
    # First, query the scipy.spatial.cKDTree (i.e. spp._kd_tree) for
    #nearest-neigh pairs
    pairs = spp._get_mating_pairs(choose_nearest=choose_nearest,
                                  inverse_dist_mating=inverse_dist_mating)
    ####################################################
    # Then, operationalize sexes, if being used, and find all 
    #available pairs within max distance
    if sex:
        # np.array of the sexes of all individuals
        sexes = np.array([ind.sex for ind in spp.values()])
        # array of couplings for all females with 
        #nearest individual < mating_radius
        # i.e.  [AT LEAST 1 INDIVID < mating_radius (OTHERWISE 
        #scipy.spatial.cKDTree associates them with a nonexistent 
        #neighbour and infinite distance if choose_nearest is True,
        #otherwise they're associated with themselves at distance==0)]
        #& [female]
        available_females =np.array(sexes[pairs[:,0]] == 0)
        # check that nearest individuals to paired females are males,
        #otherwise remove
        mating_pairs = pairs[available_females][sexes[pairs[:,
                                            1][available_females]] == 1]
    else:
        # delete inverse-equal pairs from list (i.e. if 3-paired-to-5 and 
        #5-paired-to-3 are both listed, drop one)
        # NOTE: only if mating_radius is not None, because if it is None then
        #       WF-style panmixia should be enforced, in which case we should
        #       allow for draws with replacement to generate duplicate pairs
        if spp.mating_radius is not None:
            mating_pairs = np.array(list(map(tuple, set(map(frozenset, pairs)))))
        else:
            mating_pairs = pairs
    ##############################################
    # Check if any available pairs have been found thus far, proceed if so,
    #otherwise return empty array of pairs
    if len(mating_pairs) == 0:
        mates = np.array([])
        return mates
    else:
        pass
    ###############################################
    # Then, operationalize age-structuing, if being used
    # NOTE: I have NOT restricted this so that each individual can only 
    #mate once per generation, but I should probably figure that out and 
    #add it here!
    if (repro_age is not None
        and np.any(np.atleast_1d(repro_age) > 0)):
        # np.array of the ages of all individuals
        ages = np.array([ind.age for ind in spp.values()])
        # if sexual species, repro_age expected to be a tuple or list of 
        #numerics of length 2
        if sex:
            assert repro_age.__class__.__name__ in ('ndarray', 'list',
                'tuple'), ("For a sexual and age-structured species, "
                "age at sexual maturity, 'repro_age', must be expressed as an "
                "iterable of numerics (i.e. a list, tuple, or numpy.ndarray "
                "of floats or integers); you have provided a:\n\t"
                "%s") % repro_age.__class__.__name__
            yes_f = np.array(ages[mating_pairs[:, 0]] >= repro_age[0])
            yes_m = np.array(ages[mating_pairs[:, 1]] >= repro_age[1])
            mating_pairs = mating_pairs[yes_f & yes_m]
        else:  # if not sexual species, repro_age expected to be a numeric
            assert repro_age.__class__.__name__ in ('int',
                'float'), ("For a non-sexual and age-structured species, "
                "the age at sexual maurity, 'repro_age', must be expressed as "
                "a single, non-iterable numeric (i.e.  float or integer); you "
                "have provided a:\n\t%s") % repro_age.__class__.__name__
            yes_0 = np.array(ages[mating_pairs[:, 0]] >= repro_age)
            yes_1 = np.array(ages[mating_pairs[:, 1]] >= repro_age)
            yes = np.sum(ages[mating_pairs] >= repro_age, axis = 1) == 2
            mating_pairs = mating_pairs[yes]
    #########################################
    # Then, if at least one valid pair was found, match pairs' indices
    # (in the data structure returned from get_mating_pairs)
    # back to their individual ids
    if len(mating_pairs) > 0:
        # finally, link individuals' ordinal indices back to the initially
        #created structure, to get individuals' proper keys
        f = ig(*mating_pairs.flatten())
        mates = np.array(f([*spp])).reshape(mating_pairs.shape)
    else:
        mates = np.array([])
    # Return an array or arrays, each inner array containing a mating-pair
    return mates


def _draw_n_births(num_pairs, n_births_distr_lambda):
    #NOTE: subtracting nearly 1 from the lambda, then adding 1 to the drawn
    #value, guarantees that at least 1 offspring will be born for each pair
    #chosen to mate
    num_births = r.poisson((n_births_distr_lambda), num_pairs)
    num_births = np.clip(num_births, a_min = 1, a_max = None)
    return num_births


# function for mating for a single offspring for a certain mating pair
def _do_mating_sngl_offspr(spp, pair, recomb_keys):
    # choose a start homologue
    # NOTE: for now, this will only work for diploidy
    start_homologues = r.binomial(1, 0.5, 2)

    # if using tskit, generate the segment info for each parent's new,
    # recombined gamete
    # NOTE: reverse the order of the recomb_paths object so that the resulting
    # segment info objects' order matches the order in which the recombination
    # paths are popped in the following line of code (because they will pop
    # off the right end of their list)
    if spp.gen_arch.use_tskit:
        seg_info = [spp.gen_arch.recombinations._get_seg_info(
                        start_homologue=start_homologues[i],
                        event_key=k,
                        node_ids=np.array([spp[pair[i]]._nodes_tab_ids[
                             homol] for homol in range(
                             spp.gen_arch.x)])
                        ) for i, k in enumerate(recomb_keys)]
    else:
        seg_info = None

    # generate each parent's new, recombined gamete (and make into a new genome
    # by stacking and transposing)
    # NOTE: the first path winds up the left side of the new genome, thus
    # homologue 0; this ensures that the first segment-info object in the
    # `seg_info` list and the first (i.e. left) half of the new genome
    # both correspond to the genomic material inherited from the first of the
    # two parents whose ids are listed in `pair`
    if ((spp.gen_arch.use_tskit and len(spp.gen_arch.nonneut_loci) > 0)
        or not spp.gen_arch.use_tskit):
        subsetters = [spp.gen_arch.recombinations._get_subsetter(
                                        event_key=k) for k in recomb_keys]
        #NOTE: flip the genome L-R before subsetting, if the start homologue is
        # 1, then flatten genome and subset
        new_genome = [np.fliplr(spp[ind].g).flatten(
            )[np.bool8([*sub])] if hom else spp[ind].g.flatten(
            )[np.bool8([*sub])] for ind, hom, sub in zip(pair,
                                               start_homologues, subsetters)]
        new_genome = np.vstack(new_genome).T
    else:
        new_genome = None
    return new_genome, seg_info


# function for mating a chosen mating-pair
def _do_mating_sngl_pair(spp, pair, n_births, pairs_recomb_keys):
    # generate a list of n offspring for the given pair
    offspring = [_do_mating_sngl_offspr(spp, pair,
                                        [pairs_recomb_keys.pop(
                                        ) for _ in range(2)]) for off in range(
                                                                  n_births)]
    return offspring


# function for mating a whole species
def _do_mating(spp, mating_pairs, n_births, recomb_keys):
    """
    NOT INTENDED FOR PUBLIC USE

    spp: a gnx Species object

    mating_pairs: an nx2 2d np.array of ints, where n = number of mating pairs,
                  2 columns correspond to each of the two mates in each pair,
                  and each int is the index number of an Individual in a pair

    n_births: a length-n 1d np.array of ints, where n = number of mating pairs
              and each int is the number of offspring the nth pair will produce

    recomb_keys: an m-length list of ints, where m = twice the
                 total number of births and each int is a randomly drawn
                 recombination event's key
    """
    # use list of number of offspring per mating pair to create
    # list of start and stop indices for subsetting the recomb paths into
    # groups of paths to be used for each gamete-production event for each pair
    start_stop_idxs = np.hstack((0, np.cumsum([2*n for n in n_births])))
    # get the recombination paths to be used by each pair
    pairs_recomb_keys = [recomb_keys[start_stop_idxs[i]: start_stop_idxs[
                                    i + 1]] for i in range(len(n_births))]
    # use the pairs' paths to produce recombinant genomes for each pair (where
    # number of genomes equals that pair's number of offspring)
    new_genomes = list(starmap(_do_mating_sngl_pair, zip(repeat(spp),
                               mating_pairs, n_births, pairs_recomb_keys)))
    return new_genomes



#!/usr/bin/python
# io.py

'''
Defines core IO functions
'''

# genomics imports

# other imports
import os
import csv
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
import rasterio


######################################
# -----------------------------------#
# VARIABLES -------------------------#
# -----------------------------------#
######################################


# TODO: For now, defining a default projection for saving geotiff rasters
# when no projection is set, but I need to decide if it even makes sense to do
# this, rather than just disallowing geotiff as a raster-saving format if no
# projection is set (because of course this makes random layers project as
# rasters at lat:0, lon:0 in Web Mercator, which is at the Equator and on the
# prime meridian (basically due west of São Tomé and due south of Accra)
_DEFAULT_PROJ = '''PROJCS["WGS 84 / Pseudo-Mercator",
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.0174532925199433,
            AUTHORITY["EPSG","9122"]],
        AUTHORITY["EPSG","4326"]],
    PROJECTION["Mercator_1SP"],
    PARAMETER["central_meridian",0],
    PARAMETER["scale_factor",1],
    PARAMETER["false_easting",0],
    PARAMETER["false_northing",0],
    UNIT["metre",1,
        AUTHORITY["EPSG","9001"]],
    AXIS["X",EAST],
    AXIS["Y",NORTH],
    EXTENSION["PROJ4","+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext +no_defs"],
    AUTHORITY["EPSG","3857"]]'''

######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################


#    ########
#    # Read #
#    ########

def _read_raster(filepath, coord_prec, dim=None):
    if os.path.splitext(filepath)[1].lower() == '.txt':
        rast = np.fromfile(filepath, sep=' ')
        assert len(rast) == np.prod(dim), ('The raster read in from the .txt '
                                           'file provided does not have a '
                                           'size equal to the product of the '
                                           'dimensions provided, and thus '
                                           'cannot be coerced to an array of '
                                           'those dimensions.')
        rast = np.float32(rast.reshape(dim))
        dim = rast.shape[::-1]
        res = (1, 1)
        ulc = (0, 0)
        prj = None
    else:
        rast_file = rasterio.open(filepath)
        rast = rast_file.read()[0, :, :]
        dim = rast.shape[::-1]
        # NOTE: rasterio has switched to using the proper affine transform
        # matrix as the dataset.transform attribute (see, for example,
        # https://www.perrygeo.com/python-affine-transforms.html). However, the
        # get_transform() method still returns the affine transform's tuple
        # in GDAL format, which is specified (according to the
        # tutorial at https://www.gdal.org/gdal_tutorial.html) as:
        # adfGeoTransform[0] /* top left x */
        # adfGeoTransform[1] /* w-e pixel resolution */
        # adfGeoTransform[2] /* 0 */
        # adfGeoTransform[3] /* top left y */
        # adfGeoTransform[4] /* 0 */
        # adfGeoTransform[5] /* n-s pixel resolution (negative value) */
        res = tuple([i for n, i in enumerate(rast_file.get_transform(
                                                        )) if n in [1, 5]])
        res = np.round(res, coord_prec)
        ulc = tuple([i for n, i in enumerate(rast_file.get_transform(
                                                        )) if n in [0, 3]])
        ulc = np.round(ulc, coord_prec)
        # get the projection as a CRS object
        prj = rast_file.crs
    return(rast, dim, res, ulc, prj)


# read a txt file containing a stack of 2d arrays (such as is created to save 
# LD data)
def _read_array_stack(filepath):
    array = np.loadtxt(filepath)
    return(array)

    #########
    # Write #
    #########


# write data (a block of text, as a string) to a file
def _write_file(filepath, data):
    with open(filepath, 'w') as f:
        f.write(data)


# write a CSV of data from a dictionary of columns
def _write_dict_to_csv(filepath, array_1d_dict):
    # make a pandas DataFrame from the species' stats
    df = pd.DataFrame.from_dict(array_1d_dict)
    # add a timestep column
    df['t'] = [*df.index]
    # reorder the columns so that timestep is the first
    ordered_cols = ['t'] + [col for col in df.columns if col != 't']
    df = df[ordered_cols]
    # write to disk
    df.to_csv(path_or_buf=filepath, float_format='%0.5f', index=False)


# append a 2d array layer to an array stack (i.e. an np.txt file
# intended to eventually be read in as a 3D array)
def _append_array2d_to_array_stack(filepath, locuswise_array2d):
    tmp_file_dir = os.path.split(filepath)[0]
    tmp_filename = 'tmp_%s.txt' % str(np.random.randint(0, 1000)).zfill(4)
    tmp_filename = os.path.join(tmp_file_dir, tmp_filename)
    np.savetxt(tmp_filename, locuswise_array2d, fmt='%0.5f')
    with open(tmp_filename, 'r') as f:
        tmp_txt = f.read()
    with open(filepath, 'a') as f:
        f.write(tmp_txt)
    os.remove(tmp_filename)


# append a row of data to a CSV file
def _append_row_to_csv(filepath, locuswise_array1d, t):
    # NOTE: if mean heterozygosity is requested then isntead of an array just a
    # single, floating-point value will be returned, so we then to embed that
    # inside an interable
    if (not isinstance(locuswise_array1d, list) and
        not isinstance(locuswise_array1d, np.ndarray)):
        assert (isinstance(locuswise_array1d, float) or
                isinstance(locuswise_array1d, int))
        locuswise_array1d = np.array([locuswise_array1d])
    data_dict = dict(zip(range(locuswise_array1d.size), locuswise_array1d))
    write_header = not os.path.exists(filepath)
    with open(filepath, 'a') as f:
        dict_writer = csv.DictWriter(f, ['t'] + [*data_dict.keys()])
        if write_header:
            dict_writer.writeheader()
        dict_writer.writerow({'t': t, **data_dict})


# write a data file from a geopandas object created from an index-keyed dict
# of individual.Individual objects
def _write_geopandas(filepath, individuals, driver):
    # get full path and filename
    attributes = ['idx', 'z', 'e', 'age', 'sex']
    # TODO: FIXME: replace the call to str() below with something more
    # sophisticated that will actually separate distinct phenotype
    # and environment values into separate, labeled columns
    # TODO: also round numbers to 5 decimals or so?
    df_dict = {att: [str(getattr(
            ind, att)) for ind in individuals.values()] for att in attributes}
    pts = [Point(ind.x, ind.y) for ind in individuals.values()]
    df_dict['pt'] = pts
    df = pd.DataFrame.from_dict(df_dict)
    gdf = gpd.GeoDataFrame(df, geometry='pt')
    if driver == 'CSV':
        gdf['x'] = gdf.pt.x
        gdf['y'] = gdf.pt.y
        gdf = gdf.drop(labels=('pt'), axis=1)
        gdf.to_csv(filepath, index=False)
    else:
        gdf.to_file(filepath, driver=driver)


# write a shapefile from an index-keyed dict of individual.Individual objects
def _write_shapefile(filepath, individuals):
    filepath = _set_extension(filepath, 'shp')
    _write_geopandas(filepath, individuals, driver='ESRI Shapefile')


# write a geojson from an index-keyed dict of individual.Individual objects
def _write_geojson(filepath, individuals):
    filepath = _set_extension(filepath, ['json', 'geojson'])
    _write_geopandas(filepath, individuals, driver='GeoJSON')


# write a csv from an index-keyed dict of individual.Individual objects
def _write_csv(filepath, individuals):
    filepath = _set_extension(filepath, 'csv')
    _write_geopandas(filepath, individuals, driver='CSV')


# write a txt array from a landscape.Layer object's numpy-array raster
def _write_txt_array(filepath, lyr):
    filepath = _set_extension(filepath, 'txt')
    np.savetxt(filepath, lyr.rast, fmt='%0.5f')


# write a geotiff from a landscape.Layer object's numpy-array raster
def _write_geotiff(filepath, lyr):
    filepath = _set_extension(filepath, ['tif', 'tiff'])
    # create a dict that contains all the necessary outfile metadata
    out_meta = {'driver': 'GTiff',
                'dtype': 'float32',     # always save as continuous data
                'nodata': None,
                'width': lyr.dim[0],
                'height': lyr.dim[1],   
                'count': 1,             # n layers
                'tiled': False,
                'compress': 'lzw',
                'interleave': 'band'
               }

    # take the Layer's projection, if it has one, or else just assign
    # it unprojected lat-lon, WGS84
    if lyr.prj is None:
        out_meta['crs'] = rasterio.crs.CRS.from_dict(init='epsg:4326')
    else:
        out_meta['crs'] = lyr.prj

    # use the res and ulc info to create the transform object
    aff = rasterio.transform.Affine(res[0],
                                    0,
                                    ulc[0],
                                    0,
                                    res[1],
                                    ulc[1]
                                   )
    out_meta['transform'] = aff

    with rasterio.open(filepath, 'w', **out_meta) as dst:
        dst.write(lyr.rast, 1)
    return


#    #########
#    # Other #
#    #########

# add the correct extension to a filepath, if necessary
def _set_extension(filepath, ft_ext):
    # make ft_ext iterable, if just a string is provided
    if type(ft_ext) is str:
        ft_ext = [ft_ext]

    # check if filepath has any extension
    any_ext = len([i for i in os.path.splitext(filepath) if len(i) > 0]) > 1
    # and check for an existing valid extension
    valid_ext = os.path.splitext(filepath.lower())[1].lstrip('.') in ft_ext

    # if the file has an extension not concordant with the filetype extension
    # (ft_ext), throw an error
    if any_ext and not valid_ext:
        raise ValueError(("File name already contains an extension "
                          "('%s'), but it is incompatible with the filetype "
                          "to be written (which requires one of the following "
                          " extensions: %s).") % (os.path.splitext(
                            filepath)[1], ','.join(
                            ["'.%s'" % ext for ext in ft_ext])))

    # else if it has no extension, append the extension
    elif not any_ext and not valid_ext:
        filepath = '.'.join([filepath, ft_ext[0]])
    # else, leave it be
    else:
        pass
    # return the resulting filepath
    return filepath



#!/usr/bin/python

import numpy as np
import statsmodels.api as sm
from collections import OrderedDict as OD

def MMRR(Y, X, Xnames=None, nperm=999):
    """
    This is a port of Ian Wang's MMRR script, which lives here:
        https://nature.berkeley.edu/wanglab/data/

    MMRR performs Multiple Matrix Regression with Randomization analysis

    Parameters
    ----------

    Y: {2d numpy array}
       A dependent distance matrix

    X: {[2d numpy array, 2d numpy array, ...]}
        A list of independent distance matrices

    Xnames: {[str, str, ...]}, optional
        A list of variable names for the X matrices (defaults to ["X1", "X2",
        ...])

    nperm: int, optional
       The number of permutations to use for the permutation test
    """
    # get number Y rows, then unfold lower triangular
    nrowsY = Y.shape[0]
    y = _unfold_tril(Y)
    # process X names, then unfold all Xs' lower triangulars
    if Xnames is None:
        Xnames = ["X%i" % i for i in range(1, len(X)+1)]
    xs = [_unfold_tril(x) for x in X]
    # get in correct array formats for OLS
    mod_y, mod_x = _prep_mod_data(y, xs)
    # fit the linear regression and get the stats
    mod = sm.OLS(mod_y, mod_x).fit()
    coeffs = mod.params
    r2 = mod.rsquared
    tstat = mod.tvalues
    Fstat = mod.fvalue
    tprob = np.ones((len(tstat)))
    Fprob = 1
    # get the row numbers
    rownums = [*range(nrowsY)]
    # perform permutations
    for i in range(nperm):
        # shuffle the row numbers
        np.random.shuffle(rownums)
        Yperm = Y[rownums,:][:, rownums]
        yperm = _unfold_tril(Yperm)
        permmod_y, permmod_x = _prep_mod_data(yperm, xs)
        permmod = sm.OLS(permmod_y, permmod_x).fit()
        tprob += (np.abs(permmod.tvalues) >= np.abs(tstat))
        Fprob += (permmod.fvalue >= Fstat)

    # calculate the empirical p-values
    tp = tprob/(nperm+1)
    Fp = Fprob/(nperm+1)

    # return values
    coeff_names = ["Intercept"] + Xnames
    output = OD()
    output["R^2"] = r2
    output.update({c: cval for c, cval in zip(coeff_names, coeffs)})
    output.update({c+ "(t)": tval for c, tval in zip(coeff_names, tstat)})
    output.update({c+ "(p)": pval for c, pval in zip(coeff_names, tp)})
    output["F-statistic"] = Fstat
    output["F p-value"] = Fp

    return output


def _prep_mod_data(y, xs, intercept=True):
    """
    put unfolded y and x matrix tril data into correct shapes
    """
    y_out = y.reshape(len(y), 1)
    x_out = np.vstack(xs).T
    if intercept:
        x_out = np.hstack([np.ones(y_out.shape), x_out])
    return y_out, x_out


def _unfold_tril(A):
    """
    unfolds the lower triangular elements of a matrix into a vector (i.e. 1d
    array)
    """
    vec = np.tril(A, k=-1)[np.tril_indices(A.shape[0], k=-1)]
    return vec



# biogeo_selection_example_params.py

# This is a parameters file generated by Geonomics
# (by the gnx.make_parameters_file() function).


                   #   ::::::          :::    :: :::::::::::#
             #::::::    ::::   :::      ::    :: :: ::::::::::: ::#
          #:::::::::     ::            ::   :::::::::::::::::::::::::#
        #::::::::::                      :::::::::: :::::: ::::::::  ::#
      #  : ::::  ::                    ::::  : ::    :::::::: : ::  :    #
     # GGGGG :EEEE: OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS #
    # GG     EE    OO   OO  NNN  NN  OO   OO  MM   MM   II   CC     SS     #
    # GG     EE   OO     OO NN N NN OO     OO MMM MMM   II   CC     SSSSSS #
    # GG GGG EEEE OO     OO NN  NNN OO     OO MM M MM   II   CC         SS #
    # GG   G EE    OO   OO  NN   NN  OO   OO  MM   MM   II   CC        SSS #
     # GGGGG :EEEE: OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS #
      #    : ::::::::               :::::::::: ::              ::  :   : #
        #:    :::::                    :::::: :::             :::::::  #
          #    :::                      :::::  ::              ::::: #
             #  ::                      ::::                      #
                   #                                        #
                      #  :: ::    :::             #


params = {
#-----------------------------------------------------------------------------#

#-----------------#
#--- LANDSCAPE ---#
#-----------------#
    'landscape': {

    #------------#
    #--- main ---#
    #------------#
        'main': {
            #x,y (a.k.a. j,i) dimensions of the Landscape
            'dim':                      (35, 35),
            #x,y resolution of the Landscape
            'res':                      (1,1),
            #x,y coords of upper-left corner of the Landscape
            'ulc':                      (0,0),
            #projection of the Landscape
            'prj':                      None,
            }, # <END> 'main'

    #--------------#
    #--- layers ---#
    #--------------#
        'layers': {

            #layer name (LAYER NAMES MUST BE UNIQUE!)
            'lyr_0': {

        #-------------------------------------#
        #--- layer num. 0: init parameters ---#
        #-------------------------------------#

                #initiating parameters for this layer
                'init': {

                    #parameters for an 'nlmpy'-type Layer
                    'nlmpy': {
                        #nlmpy function to use the create this Layer
                        'function':                 'mpd',
                        #number of rows (MUST EQUAL LANDSCAPE DIMENSION y!)
                        'nRow':                     35,
                        #number of cols (MUST EQUAL LANDSCAPE DIMENSION x!)
                        'nCol':                     35,
                        #level of spatial autocorrelation in element values
                        'h':                        0.7,

                        }, # <END> 'nlmpy'

                    }, # <END> 'init'

                }, # <END> layer num. 0

            'lyr_1': {

        #-------------------------------------#
        #--- layer num. 1: init parameters ---#
        #-------------------------------------#

                #initiating parameters for this layer
                'init': {

                    #parameters for an 'nlmpy'-type Layer
                    'nlmpy': {
                        #nlmpy function to use the create this Layer
                        'function':                 'mpd',
                        #number of rows (MUST EQUAL LANDSCAPE DIMENSION y!)
                        'nRow':                     35,
                        #number of cols (MUST EQUAL LANDSCAPE DIMENSION x!)
                        'nCol':                     35,
                        #level of spatial autocorrelation in element values
                        'h':                        0.7,

                        }, # <END> 'nlmpy'

                    }, # <END> 'init'

                }, # <END> layer num. 1


    #### NOTE: Individual Layers' sections can be copy-and-pasted (and
    #### assigned distinct keys and names), to create additional Layers.


            } # <END> 'layers'

        }, # <END> 'landscape'


#-----------------------------------------------------------------------------#

#-----------------#
#--- COMMUNITY ---#
#-----------------#
    'comm': {

        'species': {

            #species name (SPECIES NAMES MUST BE UNIQUE!)
            'spp_0': {

            #-----------------------------------#
            #--- spp num. 0: init parameters ---#
            #-----------------------------------#

                'init': {
                    #starting number of individs
                    'N':                500,
                    #carrying-capacity Layer name
                    'K_layer':          'lyr_0',
                    #multiplicative factor for carrying-capacity layer
                    'K_factor':         1,
                    }, # <END> 'init'

            #-------------------------------------#
            #--- spp num. 0: mating parameters ---#
            #-------------------------------------#

                'mating'    : {
                    #age(s) at sexual maturity (if tuple, female first)
                    'repro_age':                0,
                    #whether to assign sexes
                    'sex':                      False,
                    #ratio of males to females
                    'sex_ratio':                1/1,
                    #intrinsic growth rate
                    'R':                        0.5,
                    #intrinsic birth rate (MUST BE 0<=b<=1)
                    'b':                        0.2,
                    #expectation of distr of n offspring per mating pair
                    'n_births_distr_lambda':    1,
                    #whether n births should be fixed at n_births_dist_lambda
                    'n_births_fixed':           True,
                    #radius of mate-search area (None, for panmixia)
                    'mating_radius':            5,
                    #whether individs should choose nearest neighs as mates
                    'choose_nearest_mate':        False,
                    #whether mate-choice should be inverse distance-weighted
                    'inverse_dist_mating':      False,
                    }, # <END> 'mating'

            #----------------------------------------#
            #--- spp num. 0: mortality parameters ---#
            #----------------------------------------#

                'mortality'     : {
                    #maximum age
                    'max_age':                      None,
                    #min P(death) (MUST BE 0<=d_min<=1)
                    'd_min':                        0,
                    #max P(death) (MUST BE 0<=d_max<=1)
                    'd_max':                        1,
                    #width of window used to estimate local pop density
                    'density_grid_window_width':    None,
                    }, # <END> 'mortality'

            #---------------------------------------#
            #--- spp num. 0: movement parameters ---#
            #---------------------------------------#

                'movement': {
                    #whether or not the species is mobile
                    'move':                                 True,
                    #mode of distr of movement direction
                    'direction_distr_mu':                   0,
                    #concentration of distr of movement direction
                    'direction_distr_kappa':                0,
                    #1st param of distr of movement distance
                    'movement_distance_distr_param1':       0.35,
                    #2nd param of distr of movement distance
                    'movement_distance_distr_param2':       2,
                    #movement distance distr to use ('lognormal','levy','wald')
                    'movement_distance_distr':              'wald',
                    #1st param of distr of dispersal distance
                    'dispersal_distance_distr_param1':      -1,
                    #2nd param of distr of dispersal distance
                    'dispersal_distance_distr_param2':      0.05,
                    #dispersal distance distr to use ('lognormal','levy','wald')
                    'dispersal_distance_distr':             'lognormal',
                    'move_surf'     : {
                        #move-surf Layer name
                        'layer':                'lyr_0',
                        #whether to use mixture distrs
                        'mixture':              True,
                        #concentration of distrs
                        'vm_distr_kappa':       12,
                        #length of approximation vectors for distrs
                        'approx_len':           5000,
                        }, # <END> 'move_surf'

                    },    # <END> 'movement'


            #---------------------------------------------------#
            #--- spp num. 0: genomic architecture parameters ---#
            #---------------------------------------------------#

                'gen_arch': {
                    #file defining custom genomic arch
                    'gen_arch_file':            None,
                    #num of loci
                    'L':                        10,
                    #fixed starting allele freq; None/False -> rand; True -> 0.5
                    'start_p_fixed':            0.5,
                    #whether to start neutral locus freqs at 0
                    'start_neut_zero':          False,
                    #genome-wide per-base neutral mut rate (0 to disable)
                    'mu_neut':                  0,
                    #genome-wide per-base deleterious mut rate (0 to disable)
                    'mu_delet':                 0,
                    #shape of distr of deleterious effect sizes
                    'delet_alpha_distr_shape':  0.2,
                    #scale of distr of deleterious effect sizes
                    'delet_alpha_distr_scale':  0.2,
                    #alpha of distr of recomb rates
                    'r_distr_alpha':            0.5,
                    #beta of distr of recomb rates
                    'r_distr_beta':             None,
                    #whether loci should be dominant (for allele '1')
                    'dom':                      False,
                    #whether to allow pleiotropy
                    'pleiotropy':               False,
                    #custom fn for drawing recomb rates
                    'recomb_rate_custom_fn':    None,
                    #number of recomb paths to hold in memory
                    'n_recomb_paths_mem':       int(1e4),
                    #total number of recomb paths to simulate
                    'n_recomb_paths_tot':       int(1e5),
                    #num of crossing-over events (i.e. recombs) to simulate
                    'n_recomb_sims':            10_000,
                    #whether to generate recombination paths at each timestep
                    'allow_ad_hoc_recomb':      False,
                    #whether to jitter recomb bps, to correctly track num_trees
                    'jitter_breakpoints':       False,
                    #whether to save mutation logs
                    'mut_log':                  False,
                    #whether to use tskit (to record full spatial pedigree)
                    'use_tskit':                True,
                    #time step interval for simplication of tskit tables
                    'tskit_simp_interval':      100,


                    'traits': {

                        #-------------------------#
                        #---trait 0 parameters ---#
                        #-------------------------#
                        #trait name (TRAIT NAMES MUST BE UNIQUE!)
                        'trait_0': {
                            #trait-selection Layer name
                            'layer':                'lyr_1',
                            #phenotypic selection coefficient
                            'phi':                  0.6,
                            #number of loci underlying trait
                            'n_loci':               4,
                            #mutation rate at loci underlying trait
                            'mu':                   0,
                            #mean of distr of effect sizes
                            'alpha_distr_mu' :      0.25,
                            #variance of distr of effect size
                            'alpha_distr_sigma':    0,
                            #max allowed magnitude for an alpha value
                            'max_alpha_mag':        None,
                            #curvature of fitness function
                            'gamma':                1,
                            #whether the trait is universally advantageous
                            'univ_adv':             False
                            }, # <END> trait 0


    #### NOTE: Individual Traits' sections can be copy-and-pasted (and
    #### assigned distinct keys and names), to create additional Traits.


                        }, # <END> 'traits'

                    }, # <END> 'gen_arch'


                }, # <END> spp num. 0



    #### NOTE: individual Species' sections can be copy-and-pasted (and
    #### assigned distinct keys and names), to create additional Species.


            }, # <END> 'species'

        }, # <END> 'comm'


#-----------------------------------------------------------------------------#

#-------------#
#--- MODEL ---#
#-------------#
    'model': {
        #total Model runtime (in timesteps)
        'T':            100,
        #min burn-in runtime (in timesteps)
        'burn_T':       30,
        #seed number
        'num':          None,


        #-----------------------------#
        #--- iterations parameters ---#
        #-----------------------------#
        'its': {
            #num iterations
            'n_its':            1,
            #whether to randomize Landscape each iteration
            'rand_landscape':   False,
            #whether to randomize Community each iteration
            'rand_comm':        False,
            #whether to randomize GenomicArchitectures each iteration
            'rand_genarch':     True,
            #whether to burn in each iteration
            'repeat_burn':      False,
            }, # <END> 'iterations'



        } # <END> 'model'

    } # <END> params
